/* Particle generators for the sandbox.
   Copyright (C) 2006 Russell Turner,
                      Thomas Ren Sidor,
                      Carl Fredrik Hammar (hammy.lite@gmail.com)

   This file is part of GP2X Sand.

   GP2X Sand is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   GP2X Sand is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details. 

   You should have received a copy of the GNU General Public License along
   with GP2X Sand; if not, write to the Free Software Foundation, Inc., 51
   Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA */

#include "emitter.h"

#include <assert.h>
#include <stdlib.h>

#include "pen.h"
#include "utils.h"

/* An emitter sits at the top of the sandbox generating new particles each
   frame.  The rate it generates new particles depends on it's width and
   it's flow, the probability of a particle getting generated at each
   pixel directly beneath the emitter.

   Emitters have the following attributes:
   Element:      The type of particle to emit.

   Flow:         The probability that a particle gets generated at each
                 of the pixels directly beneath the emitter. Specified as
                 a value between 0 and 1.

   Position:     At which pixel at the top of the sandbox the emitter
                 begins.

   Width:        Width of the emitter.

   Enabled flag: Whether the emitter should emit new particles. */

/* An action that also stores by how much the flow should be changed.  For
   use by actions changing the flow of the emitter. */
struct emitter_action_str
{
  action_t action;
  float flow_step;
};

typedef struct emitter_action_str emitter_action_t;
  
/* The three default emitters, that emits the element suggested by there
   name. 
  
   FIXME:  This should not be hard-coded and should probably be moved to the
   sandbox module. */
emitter_t emitter_sand;
emitter_t emitter_water;
emitter_t emitter_oil;
emitter_t emitter_salt;
		       
/* Forward declarations. */
void
emitters_actions_init ();

emitter_t*
emitter_selected_get ();

void
emitter_selected_action_enabled_toggle_function (void *action_ptr,
						 SDL_Event *event_ptr);

void
emitter_selected_action_flow_step_init (emitter_action_t *action_ptr,
					float flow_step);

void
emitter_selected_action_flow_function (void *action_ptr,
				       SDL_Event *event_ptr);

/* Returns FLOW if it's within the interval 0 to 1, otherwise it returns
   either 0 or 1 depending on which is closest to FLOW. */
float
flow_normalize (float flow)
{
  return clip_to_intervalf (0.00f, 1.0f, flow);
}

/* Initialize the emitter pointed to by EMITTER_PTR to have the specified
   properties.  EMITTER_PTR mustn't be null, if FLOW isn't between 0 to 1
   it it's set to the closest value that is.*/
void
emitter_init (emitter_t *emitter_ptr, element_t *element_ptr, float flow,
	      int position, int width, bool enabled_flag)
{
  assert (emitter_ptr != 0);
  assert (element_ptr != 0);
  
  emitter_ptr->element_ptr = element_ptr;
  emitter_ptr->flow = flow_normalize (flow);
  emitter_ptr->position = position;
  emitter_ptr->width = width;
  emitter_ptr->enabled_flag = enabled_flag;

  emitters_actions_init ();
}
  
/* Initialize every property to zero or false.  EMITTER_PTR mustn't be
   null. */
void
emitter_init_default (emitter_t *emitter_ptr)
{
  emitter_init (emitter_ptr, element_nothing_ptr, 0.0f, 0, 0, false);
}

/* Generate new particles directly beneath the emitter.  Whether a
   particle gets generated at a pixel depends on the flow of the emitter
   which is the probability of this happening. EMITTER_PTR mustn't be
   null. */ 
void
emitter_emit (emitter_t *emitter_ptr)
{
  assert (emitter_ptr != 0);

  if (!emitter_ptr->enabled_flag)
    return;
  
  for (int x = emitter_ptr->position - emitter_ptr->width / 2;
       x < emitter_ptr->position + emitter_ptr->width / 2;
       x++)
    {
      if (rand() < (int) (RAND_MAX * emitter_ptr->flow))
	sandbox_particle_init (vcons (x, 1), emitter_ptr->element_ptr);
    }
}

/* Enables the emitter if it was disabled and returns true or disable it
   if enabled and returns false.  EMITTER_PTR mustn't be null. */
bool
emitter_enabled_toggle (emitter_t *emitter_ptr)
{
  assert (emitter_ptr != 0);

  return emitter_ptr->enabled_flag = !emitter_ptr->enabled_flag;
}

/* Change the flow of the emitter by STEP and returns the new flow of the
   emitter.  The flow won't get outside of the range 0 to 1.  EMITTER_PTR
   mustn't be null. */
float
emitter_flow_step (emitter_t *emitter_ptr, float step)
{
  assert (emitter_ptr != 0);

  emitter_ptr->flow = flow_normalize (emitter_ptr->flow + step);
  return emitter_ptr->flow;
}

/** Emitter actions. **/

/* The selected is the emitter of the same element that the pen currently
   paints with, or nothing if there is no corresponding emitter. */

/* Toggle the selected emitter between on and off */
action_t emitter_selected_action_enabled_toggle;

/* Increase the flow of the selected emitter by 0.05 up to 1. */
emitter_action_t emitter_selected_action_flow_increase;

/* Decrease the flow of the selected emitter by 0.05 down to 0. */
emitter_action_t emitter_selected_action_flow_decrease;

/* Initiates the actions pointers. */
action_t *const emitter_selected_action_enabled_toggle_ptr =
  (void*) &emitter_selected_action_enabled_toggle;
action_t *const emitter_selected_action_flow_increase_ptr =
  (void*) &emitter_selected_action_flow_increase;
action_t *const emitter_selected_action_flow_decrease_ptr =
  (void*) &emitter_selected_action_flow_decrease;

/* Initialize all the emitter actions. */
void
emitters_actions_init ()
{
  static bool init_done_flag = false;
  
  if (init_done_flag)
    return;

  float flow_step = 0.05f;
  
  action_init (&emitter_selected_action_enabled_toggle,
	       emitter_selected_action_enabled_toggle_function);

  emitter_selected_action_flow_step_init
    (&emitter_selected_action_flow_increase, flow_step);
  emitter_selected_action_flow_step_init
    (&emitter_selected_action_flow_decrease, -flow_step);
  
  init_done_flag = true;
}

/* Return the emitter which element is the same as the pens paint element
   or null if there is no corresponding emitter. */
emitter_t*
emitter_selected_get ()
{
  if (pen.paint_element_ptr == element_sand_ptr)
    return &emitter_sand;
      
  else if (pen.paint_element_ptr == element_water_ptr)
    return &emitter_water;
      
  else if (pen.paint_element_ptr == element_oil_ptr)
    return &emitter_oil;
      
  else if (pen.paint_element_ptr == element_salt_ptr)
    return &emitter_salt;
      
  else
    return 0;
}

/* Initialize a emitter action that changes the flow of the selected
   emitter by FLOW_STEP much.  ACTION_PTR mustn't be null. */
void
emitter_selected_action_flow_step_init (emitter_action_t *action_ptr,
					float flow_step)
{
  assert (action_ptr);
  
  action_init ((action_t *) action_ptr,
	       emitter_selected_action_flow_function);
  action_ptr->flow_step = flow_step;
}

/* Function assigned to emitter_selected_action_enabled_toggle. */
void
emitter_selected_action_enabled_toggle_function (void *action_ptr,
						 SDL_Event *event_ptr)
{
  if (event_trigger_state (event_ptr) == TRIGGER_DOWN)
    {
      emitter_t *emitter_ptr = emitter_selected_get ();
      
      if (emitter_ptr != 0)
	emitter_enabled_toggle (emitter_ptr);
    }
}

/* Function used by the flow changing actions.  Fetches the FLOW_STEP
   stored in the emitter action pointed to by ACTION_PTR and changes the
   flow of the selected emitter by that much. */
void
emitter_selected_action_flow_function (void *action_ptr,
				       SDL_Event *event_ptr)
{
  if (event_trigger_state (event_ptr) == TRIGGER_DOWN && action_ptr != 0)
    {
      emitter_action_t *emitter_action_ptr = action_ptr;
      emitter_t *emitter_ptr = emitter_selected_get ();
      
      if (emitter_ptr != 0)
	emitter_flow_step (emitter_ptr, emitter_action_ptr->flow_step);
    }
}
