/* Handles events from SDL.
   Copyright (C) 2006 Russell Turner,
                      Thomas Ren Sidor,
                      Carl Fredrik Hammar (hammy.lite@gmail.com)

   This file is part of GP2X Sand.

   GP2X Sand is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   GP2X Sand is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details. 

   You should have received a copy of the GNU General Public License along
   with GP2X Sand; if not, write to the Free Software Foundation, Inc., 51
   Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA */

#include "eventhandler.h"

#include <stdbool.h>
#include <assert.h>

#include "pen.h"

/* The event-handler maps events generated by the keyboard and joystick to
   actions and invokes the actions when the corresponding event occurs. */

/* Points to the action invoked on SDL_Quit events. */
action_t *action_quit_ptr;

/* Lookup table with pointers to the actions that gets invoked on
   corresponding key events. */
action_t *map_key[SDLK_LAST];

/* Lookup table with pointers to the actions that gets invoked on
   corresponding joystick button events. */
action_t *map_joybutton[256];

/* Map the SDL_Quit event to the action pointed to by ACTION_PTR. */
void
eventhandler_quit_map (action_t *action_ptr)
{
  action_quit_ptr = action_ptr;
}

/* Map event generated by KEY to the action pointed to by ACTION_PTR. 
   KEY must be a valid keyboard key, */
void
eventhandler_key_map (SDLKey key, action_t *action)
{
  map_key[key] = action;
}

/* Map all the mappings specified by the mappings in the array MAP, where
   all button entries will be viewed as keyboard keys. This is equivalent
   to calling eventhandler_key_map for each entry in MAP.  MAP_LENGTH must
   be the length of MAP. */
void
eventhandler_key_map_load (mapping_button_action_t *map,
			   size_t map_lenght)
{
  assert (map != 0);
  
  for (int i = 0; i < map_lenght; i++)
    eventhandler_key_map (map[i].button, map[i].action);
}

/* Map event generated by joystick button JOYBUTTON to the action pointed
   to by ACTION_PTR. JOYBUTTON must be between 0 and 255. */ 
void
eventhandler_joybutton_map (int joybutton, action_t *action)
{
  map_joybutton[joybutton] = action;
}
  
/* Map all the mappings specified by the mappings in the array MAP, where
   all button entries will be viewed as joystick buttons. This is
   equivalent to calling eventhandler_joybutton_map for each entry in MAP.
   MAP_LENGTH must be the length of MAP. */
void
eventhandler_joybutton_map_load (mapping_button_action_t *map,
				 size_t map_lenght)
{
  assert (map != 0);
  
  for (int i = 0; i < map_lenght; i++)
    eventhandler_joybutton_map (map[i].button, map[i].action);
}

/* Initialize the event-handler, with no mappings. */
void
eventhandler_init_default ()
{
  /* This actually did something before. No point in removing it
     though. */
}

/* Process the events in SDL's event queue and invoke corresponding
   actions.  Call this regularly. */
void
eventhandler_process_events ()
{
  SDL_Event event;

  //Polling events
  while (SDL_PollEvent(&event))
    {
      switch (event.type)
	{
	case SDL_QUIT:
	  action_invoke (action_quit_ptr, &event);
	  break;
	  
	  //Key strokes
	case SDL_KEYDOWN:
	case SDL_KEYUP:
	  action_invoke (map_key[event.key.keysym.sym], &event);
	  break;

	  // If mouse button pressed then save position of cursor
	case SDL_MOUSEBUTTONDOWN:
	  pen_draw_flag_set (true);
	  break;

	  // Button released
	case SDL_MOUSEBUTTONUP:
	  pen_draw_flag_set (false);
	  break;

	  // Mouse has moved
	case SDL_MOUSEMOTION:
	  pen_warp (event.motion.x, event.motion.y);
	  break;

	  // Joystick button pressed or released.
	case SDL_JOYBUTTONDOWN:
	case SDL_JOYBUTTONUP:
	  action_invoke (map_joybutton[event.jbutton.button], &event);
	      
	default:
	  break;
	}
    }
}
