/* Starting point for GP2X Sand.
   Copyright (C) 2006 Russell Turner,
                      Thomas Ren Sidor,
                      Carl Fredrik Hammar (hammy.lite@gmail.com)

   This file is part of GP2X Sand.

   GP2X Sand is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   GP2X Sand is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details. 

   You should have received a copy of the GNU General Public License along
   with GP2X Sand; if not, write to the Free Software Foundation, Inc., 51
   Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA */

#include <stdbool.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include "SDL.h"

#include "display.h"
#include "emitter.h"
#include "eventhandler.h"
#include "fpscounter.h"
#include "globals.h"
#include "gp2xkeyboard.h"
#include "gp2xbuttons.h"
#include "gp2xstick.h"
#include "particle-counter.h"
#include "pen.h"
#include "sandbox.h"
#include "utils.h"

SDL_Surface *screen;
bool done_flag;
  
action_t action_quit;
action_t *action_quit_ptr = &action_quit;

void
controls_map ();

void
action_quit_function (void *action_ptr, SDL_Event *event_ptr)
{
  done_flag = true;
}

// Initialize subsystems
void
init()
{
  // Initialize SDL.
  if (SDL_Init (SDL_INIT_VIDEO | SDL_INIT_JOYSTICK) < 0)
    {
      fprintf (stderr, "Video initialization failed: %s\n",
	       SDL_GetError ());
      SDL_Quit ();
    }

  // Create the screen using 16 bit colors.
  screen = SDL_SetVideoMode(SCREEN_WIDTH, SCREEN_HEIGHT, 16,
			    SDL_HWSURFACE);
  if (screen == NULL)
    fprintf (stderr, "Unable to set video mode: %s\n", SDL_GetError ());

  // Set caption.
  SDL_WM_SetCaption ("SDL Sand", NULL);

  // Open the first joystick if present.
  // The controls of a GP2X will be mapped to this.
  SDL_JoystickOpen (0);

  // Move cursor to the center of the screen.
  //SDL_WarpMouse (SCREEN_WIDTH / 2, SCREEN_HEIGHT / 2);

  // Set initial seed.
  srand ((unsigned) time (NULL));

  sandbox_display_init (screen, SCREEN_WIDTH, SCREEN_HEIGHT);
  pen_init_default ();
  sandbox_init ();
  eventhandler_init_default ();

  emitter_init (&emitter_sand, element_sand_ptr, 0.3f,
		SCREEN_WIDTH / 5, 5, true);
  emitter_init (&emitter_water, element_water_ptr, 0.3f,
		SCREEN_WIDTH / 5 * 2, 5, true);
  emitter_init (&emitter_oil, element_oil_ptr, 0.3f,
		SCREEN_WIDTH / 5 * 3, 5, true);
  emitter_init (&emitter_salt, element_salt_ptr, 0.3f,
		SCREEN_WIDTH / 5 * 4, 5, true);

  action_init (&action_quit, action_quit_function);

  fps_counter_init (SDL_GetTicks ());

  eventhandler_quit_map (action_quit_ptr);
  controls_map ();
}

void
controls_map ()
{
  mapping_button_action_t map_joybuttons[] = 
    { { GP2X_BUTTON_B,       pen_action_draw_ptr },
      { GP2X_BUTTON_START,   action_quit_ptr },
      { GP2X_BUTTON_SELECT,  sandbox_action_clear_ptr },
      { GP2X_BUTTON_Y,       pen_action_paint_element_cycle_forward_ptr },
      { GP2X_BUTTON_A,       pen_action_paint_element_cycle_backward_ptr },
      { GP2X_BUTTON_R,       pen_action_size_increase_ptr },
      { GP2X_BUTTON_L,       pen_action_size_decrease_ptr },
      { GP2X_BUTTON_X,       emitter_selected_action_enabled_toggle_ptr },
      { GP2X_BUTTON_VOLUP,   emitter_selected_action_flow_increase_ptr },
      { GP2X_BUTTON_VOLDOWN, emitter_selected_action_flow_decrease_ptr } };
  
  int map_joybuttons_n = LENGTH_OF (map_joybuttons);

  eventhandler_joybutton_map_load (map_joybuttons, map_joybuttons_n);

  stick_init_default ();
  gp2x_keyboard_init_default ();
}

void
cleanup ()
{
  SDL_Quit ();

  // Restart the menu.
  chdir ("/usr/gp2x");
  execl ("/usr/gp2x/gp2xmenu", "/usr/gp2x/gp2xmenu", NULL);
}

int
main (int argc, char **argv)
{
  init ();

  int tick = 0;

  /* Main loop. */
  while (!done_flag)
    {
      tick++;

      eventhandler_process_events ();

      stick_pen_move ();
      pen_draw ();

      /* Fire the emitters. */
      emitter_emit (&emitter_sand);
      emitter_emit (&emitter_water);
      emitter_emit (&emitter_oil);
      emitter_emit (&emitter_salt);
      
      sandbox_update ();
      sandbox_display_update ();

      SDL_Flip (screen);
	  
      fps_counter_frame_timestamp_push (SDL_GetTicks ());
      
      /* Print info every 50 frames. */
      if (tick % 50 == 0)
	{
	  /* Print the frame rate and particle count. */
	  printf ("FPS: %i\n", fps_counter_framerate_get ());
	  printf ("Particle count: %i\n", sandbox_particles_count ());
	}
    }

  cleanup ();

  return 0;
}
